#!@PYTHON_EXECUTABLE@
# ************************************************************************
#
#                        Kokkos v. 3.0
#       Copyright (2020) National Technology & Engineering
#               Solutions of Sandia, LLC (NTESS).
#
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the Corporation nor the names of the
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY NTESS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NTESS OR THE
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Contact Christian R. Trott (crtrott@sandia.gov)
#
# ************************************************************************
#


from __future__ import absolute_import

__author__ = "Jonathan R. Madsen"
__copyright__ = (
    "Copyright 2020, National Technology & Engineering Solutions of Sandia, LLC (NTESS)"
)
__credits__ = ["Kokkos"]
__license__ = "BSD-3"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan R. Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"


import os
import unittest


# discover and run all unittests in the current directory
def run_all_tests():
    # auto discover unittests from *.py files starting character into the test suite
    test_suite = unittest.defaultTestLoader.discover(
        start_dir=os.path.dirname(os.path.abspath(__file__)),
        pattern="[a-z]*.py",
    )

    # print the loaded tests
    print(f"============= Loaded Tests =============\n\n {test_suite}\n")

    # create a results object to store test results
    result = unittest.TestResult(verbosity=2)

    # run all tests in test_suite, use result object to store results
    print("\n============= Tests Stdout =============\n")
    # run the tests
    test_suite.run(result)

    # print the results
    print(f"\n============= Results =============\n{result}\n")
    return result



#!@PYTHON_EXECUTABLE@
# ************************************************************************
#
#                        Kokkos v. 3.0
#       Copyright (2020) National Technology & Engineering
#               Solutions of Sandia, LLC (NTESS).
#
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the Corporation nor the names of the
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY NTESS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NTESS OR THE
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Contact Christian R. Trott (crtrott@sandia.gov)
#
# ************************************************************************
#

from __future__ import absolute_import

__author__ = "Jonathan R. Madsen"
__copyright__ = (
    "Copyright 2020, National Technology & Engineering Solutions of Sandia, LLC (NTESS)"
)
__credits__ = ["Kokkos"]
__license__ = "BSD-3"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan R. Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"


import kokkos
import unittest


data = {}
temp = {}
count = 0
regions = []


def print_help(arg0):
    pass


def parse_args(argc, argv):
    global data

    data["parse_args"] = (argc, argv)


def initialize(seq, ver, deviceCount, deviceInfo):
    global data

    if "initialize" not in data:
        data["initialize"] = 1
    else:
        data["initialize"] += 1


def finalize():
    global data

    if "finalize" not in data:
        data["finalize"] = 1
    else:
        data["finalize"] += 1


def begin_parallel(name, devid):
    global data
    global temp
    global count

    count += 1
    kernid = count
    data[f"begin_{name}"] = kernid
    temp[kernid] = name
    return kernid


def end_parallel(kernid):
    global data
    global temp

    name = temp[kernid]
    data[f"end_{name}"] = kernid


def push_region(name):
    global data
    global regions

    data[name] = [True, False]
    regions += [name]


def pop_region():
    global data
    global regions

    data[regions[-1]][1] = True


def alloc_data(handle, label, ptr, size):
    global data

    data[f"alloc_{label}"] = (size, handle.name)


def dealloc_data(handle, label, ptr, size):
    global data

    data[f"dealloc_{label}"] = (size, handle.name)


def create_prof(name):
    global data
    global temp
    global count

    count += 1
    secid = count
    data[f"create_{name}"] = secid
    temp[secid] = name
    return secid


def start_prof(secid):
    global data
    global temp

    name = temp[secid]
    data[f"start_{name}"] = secid


def stop_prof(secid):
    global data
    global temp

    name = temp[secid]
    data[f"stop_{name}"] = secid


def destroy_prof(secid):
    global data
    global temp

    name = temp[secid]
    data[f"destroy_{name}"] = secid


def prof_event(name):
    global data

    data[f"{name}"] = True


def begin_deep_copy(dst_handle, dst_name, dst_ptr, src_handle, src_name, src_ptr, size):
    global data
    global temp

    data[f"begin_{dst_name}/{dst_handle.name}/dst"] = [size, True, False]
    data[f"begin_{src_name}/{src_handle.name}/src"] = [size, True, False]
    temp["deep_copy"] = (dst_handle, dst_name, src_handle, src_name, size)


def end_deep_copy():
    global data
    global temp

    (dst_handle, dst_name, src_handle, src_name, size) = temp["deep_copy"]
    del temp["deep_copy"]

    data[f"begin_{dst_name}/{dst_handle.name}/dst"][-1] = True
    data[f"begin_{src_name}/{src_handle.name}/src"][-1] = True


class PyKokkosBaseToolsTests(unittest.TestCase):
    """ToolsTests"""

    @classmethod
    def setUpClass(self):
        import gc

        kokkos.tools.set_parse_args_callback(parse_args)
        kokkos.tools.set_print_help_callback(print_help)
        kokkos.tools.set_init_callback(initialize)
        kokkos.tools.set_finalize_callback(finalize)

        kokkos.initialize()
        # self.assertTrue(data["initialize"])

        # configure internal testing data before setting other callbacks
        kokkos.tools._internal.setup()

        for itr in ("parallel_for", "parallel_reduce", "parallel_scan", "fence"):
            import sys

            begin_func = getattr(sys.modules[__name__], "begin_parallel")
            end_func = getattr(sys.modules[__name__], "end_parallel")
            getattr(kokkos.tools, f"set_begin_{itr}_callback")(begin_func)
            getattr(kokkos.tools, f"set_end_{itr}_callback")(end_func)

        kokkos.tools.set_push_region_callback(push_region)
        kokkos.tools.set_pop_region_callback(pop_region)
        kokkos.tools.set_allocate_data_callback(alloc_data)
        kokkos.tools.set_deallocate_data_callback(dealloc_data)
        kokkos.tools.set_create_profile_section_callback(create_prof)
        kokkos.tools.set_start_profile_section_callback(start_prof)
        kokkos.tools.set_stop_profile_section_callback(stop_prof)
        kokkos.tools.set_destroy_profile_section_callback(destroy_prof)
        kokkos.tools.set_profile_event_callback(prof_event)
        kokkos.tools.set_begin_deep_copy_callback(begin_deep_copy)
        kokkos.tools.set_end_deep_copy_callback(end_deep_copy)

        # run internal testing (parallel_for, parallel_reduce, etc.)
        kokkos.tools._internal.test()

        # disable parallel callbacks
        for itr in ("parallel_for", "parallel_reduce", "parallel_scan"):
            getattr(kokkos.tools, f"set_begin_{itr}_callback")(None)
            getattr(kokkos.tools, f"set_end_{itr}_callback")(None)

        # generate some profiling data in python
        source = kokkos.array(
            "python_source", [10], space=kokkos.Host, dtype=kokkos.int64
        )
        target = kokkos.array(
            "python_target", [10], space=kokkos.Host, dtype=kokkos.int64
        )
        kokkos.deep_copy(target, source)

        # delete the source and target and run garbage collector
        del source
        del target
        gc.collect()

    @classmethod
    def tearDownClass(self):
        global data

        print("")
        for key, item in data.items():
            print("{:30} : {}".format(key, item))

        for itr in (
            "parse_args",
            "print_help",
            "init",
            "finalize",
            "push_region",
            "pop_region",
            "allocate_data",
            "deallocate_data",
            "create_profile_section",
            "start_profile_section",
            "stop_profile_section",
            "destroy_profile_section",
            "begin_parallel_for",
            "begin_parallel_reduce",
            "begin_parallel_scan",
            "begin_fence",
            "begin_deep_copy",
            "end_parallel_for",
            "end_parallel_reduce",
            "end_parallel_scan",
            "end_fence",
            "end_deep_copy",
            "profile_event",
        ):
            getattr(kokkos.tools, f"set_{itr}_callback")(None)

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def fib(self, n):
        return n if n < 2 else self.fib(n - 1) + self.fib(n - 2)

    def test_routines(self):
        """cxx_routines"""

        self.assertEqual(data["begin_cxx_parallel_for"], 1)
        self.assertEqual(data["begin_cxx_parallel_reduce"], 2)
        self.assertEqual(data["begin_cxx_parallel_scan"], 3)

        self.assertEqual(data["end_cxx_parallel_for"], 1)
        self.assertEqual(data["end_cxx_parallel_reduce"], 2)
        self.assertEqual(data["end_cxx_parallel_scan"], 3)

        self.assertEqual(data["create_cxx_created_section"], 4)
        self.assertEqual(data["start_cxx_created_section"], 4)
        self.assertEqual(data["stop_cxx_created_section"], 4)
        self.assertEqual(data["destroy_cxx_created_section"], 4)

        self.assertEqual(data["create_python_profile_section"], 5)
        self.assertEqual(data["start_python_profile_section"], 5)
        self.assertEqual(data["stop_python_profile_section"], 5)
        self.assertEqual(data["destroy_python_profile_section"], 5)

        self.assertCountEqual(data["begin_cxx_target/Host/dst"], [40, True, True])
        self.assertCountEqual(data["begin_cxx_source/Host/src"], [40, True, True])

        self.assertCountEqual(data["cxx_push_region"], [True, True])

        self.assertCountEqual(data["alloc_python_source"], [80, "Host"])
        self.assertCountEqual(data["alloc_python_target"], [80, "Host"])
        self.assertCountEqual(data["dealloc_python_source"], [80, "Host"])
        self.assertCountEqual(data["dealloc_python_target"], [80, "Host"])
        self.assertCountEqual(data["dealloc_cxx_source"], [40, "Host"])
        self.assertCountEqual(data["dealloc_cxx_target"], [40, "Host"])

    def test_region(self):
        """python_region"""
        kokkos.tools.push_region(self.shortDescription())
        kokkos.tools.pop_region()

    def test_profile_section(self):
        """python_profile_section"""
        idx = kokkos.tools.create_profile_section(self.shortDescription())
        kokkos.tools.start_section(idx)
        self.fib(10)
        kokkos.tools.stop_section(idx)
        kokkos.tools.destroy_profile_section(idx)

    def test_mark_event(self):
        """python_mark_event"""
        kokkos.tools.mark_event(self.shortDescription())

    def test_declare_metadata(self):
        """python_declare_metadata"""
        kokkos.tools.declare_metadata("dogs", "good")


# main runner
def run():
    # run all tests
    unittest.main()


if __name__ == "__main__":
    run()



#!@PYTHON_EXECUTABLE@
# ************************************************************************
#
#                        Kokkos v. 3.0
#       Copyright (2020) National Technology & Engineering
#               Solutions of Sandia, LLC (NTESS).
#
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the Corporation nor the names of the
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY NTESS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NTESS OR THE
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Contact Christian R. Trott (crtrott@sandia.gov)
#
# ************************************************************************
#


from __future__ import absolute_import

__author__ = "Jonathan R. Madsen"
__copyright__ = (
    "Copyright 2020, National Technology & Engineering Solutions of Sandia, LLC (NTESS)"
)
__credits__ = ["Kokkos"]
__license__ = "BSD-3"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan R. Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"


def get_max_concrete_dims():
    import kokkos

    return kokkos.max_concrete_rank


def get_dtypes():
    import kokkos

    return kokkos.dtypes


def get_memory_spaces(only_available=True):
    import kokkos

    return [
        e
        for e in kokkos.memory_spaces
        if not only_available or kokkos.get_memory_space_available(e)
    ]


def get_layouts(only_available=True):
    import kokkos

    return [
        e
        for e in kokkos.layouts
        if not only_available or kokkos.get_layout_available(e)
    ]


def get_memory_traits(only_available=True):
    import kokkos

    return [
        e
        for e in kokkos.memory_traits
        if not only_available or kokkos.get_memory_trait_available(e)
    ]


def generate_variant(*_args, **_kwargs):
    """Generate a view from the variant arguments"""

    import kokkos
    import numpy as np

    con_arr = None
    dyn_arr = None
    if _kwargs["trait"] == kokkos.Unmanaged:
        con_arr = np.zeros(_args[0], dtype=kokkos.convert_dtype(_kwargs["dtype"]))
        con_view = kokkos.unmanaged_array(con_arr, **_kwargs, dynamic=False)
        dyn_arr = np.zeros(_args[0], dtype=kokkos.convert_dtype(_kwargs["dtype"]))
        dyn_view = kokkos.unmanaged_array(dyn_arr, **_kwargs, dynamic=True)
    else:
        con_view = kokkos.array(*_args, **_kwargs, dynamic=False)
        dyn_view = kokkos.array(*_args, **_kwargs, dynamic=True)
    # retain con_arr and dyn_arr since python might run GC and delete them
    return [con_view, dyn_view, con_arr, dyn_arr]


def get_variants(exclude=[]):
    """Return a list of all view variants"""
    _variants = []

    def _skip(_key, _value):
        return _key in exclude.keys() and _value in exclude[_key]

    for _dims in range(1, get_max_concrete_dims()):
        _shape = []
        _idx = []
        _zeros = []

        for i in range(_dims):
            _shape.append(2)
            _zeros.append(0)
            _idx.append((i + 1) % 2)

        for _dtype in get_dtypes():
            if _skip("dtypes", _dtype):
                continue
            for _space in get_memory_spaces():
                if _skip("memory_spaces", _space):
                    continue
                for _layout in get_layouts():
                    if _skip("layouts", _layout):
                        continue
                    for _trait in get_memory_traits():
                        if _skip("memory_traits", _trait):
                            continue
                        _variant = {}
                        _variant["dtype"] = _dtype
                        _variant["space"] = _space
                        _variant["layout"] = _layout
                        _variant["trait"] = _trait
                        _variants.append([_shape, _idx, _zeros, _variant])
    return _variants



#!@PYTHON_EXECUTABLE@
# ************************************************************************
#
#                        Kokkos v. 3.0
#       Copyright (2020) National Technology & Engineering
#               Solutions of Sandia, LLC (NTESS).
#
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the Corporation nor the names of the
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY NTESS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NTESS OR THE
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Contact Christian R. Trott (crtrott@sandia.gov)
#
# ************************************************************************
#

from __future__ import absolute_import
from . import libpykokkos as lib

__author__ = "Jonathan R. Madsen"
__copyright__ = (
    "Copyright 2020, National Technology & Engineering Solutions of Sandia, LLC (NTESS)"
)
__credits__ = ["Kokkos"]
__license__ = "BSD-3"
__version__ = "3.1.1"
__maintainer__ = "Jonathan R. Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"


def convert_dtype(_dtype, _module=None):
    """Converts kokkos data types into numpy dtype"""
    if isinstance(_dtype, (str, int)):
        _true_dtype = lib.get_dtype(_dtype)
    else:
        _true_dtype = lib.get_dtype(int(_dtype))
    if _module is None:
        import numpy as np

        _module = np
    return getattr(_module, _true_dtype)


def read_dtype(_dtype):
    """Converts generic dtype (from numpy for example) into kokkos dtype"""
    # try calling library routine
    if isinstance(_dtype, (str, int)):
        try:
            return getattr(lib, lib.get_dtype(_dtype))
        except ValueError:
            pass

    # try numpy
    try:
        import numpy as np

        if _dtype == np.int16:
            return lib.int16
        elif _dtype == np.int32:
            return lib.int32
        elif _dtype == np.int64:
            return lib.int64
        elif _dtype == np.uint16:
            return lib.uint16
        elif _dtype == np.uint32:
            return lib.uint32
        elif _dtype == np.uint64:
            return lib.uint64
        elif _dtype == np.float32:
            return lib.float32
        elif _dtype == np.float64:
            return lib.float64
    except ImportError:
        pass

    # just return the dtype
    return _dtype


def _determine_array_input(_inp, shape, label, array, dtype, layout):
    """Determine whether first argument is shape, label, or array
    and extract subsequent info"""
    # support non-labeled variant
    if isinstance(_inp, str) and label is None:
        label = f"{_inp}"
    elif isinstance(_inp, (list, tuple)) and shape is None:
        shape = list(_inp)[:]
    elif array is None:
        array = _inp
        try:
            shape = array.shape
            _order = array.order
            if _order == "C":
                layout = lib.LayoutRight
            elif _order == "F":
                layout = lib.LayoutLeft
        except AttributeError:
            pass

    # if array has been passed in, try getting the dtype
    if array is not None:
        try:
            dtype = read_dtype(array.dtype)
        except AttributeError:
            pass

    return [shape, label, array, dtype, layout]


def array(
    shape_label_or_array,
    shape=None,
    label=None,
    array=None,
    dtype=lib.double,
    space=lib.HostSpace,
    layout=lib.LayoutRight,
    trait=lib.Managed,
    dynamic=False,
    order=None,
):
    [shape, label, array, dtype, layout] = _determine_array_input(
        shape_label_or_array, shape, label, array, dtype, layout
    )

    # layout was specified via numpy "order" field
    if order is not None and layout == lib.LayoutRight and isinstance(order, str):
        if order.upper() == "C":
            layout = lib.LayoutRight
        elif order.upper() == "F":
            layout = lib.LayoutLeft

    _prefix = "KokkosView" if not dynamic else "KokkosDynRankView"
    _space = lib.get_memory_space(space)
    _dtype = lib.get_dtype(dtype)
    _layout = lib.get_layout(layout)
    _name = None
    _label = None
    _ndim = len(shape)

    if dynamic:
        _dtype_str = "{}".format(_dtype)
    else:
        if _ndim > lib.max_concrete_rank:
            raise ValueError(
                "pykokkos-base build only supports {} ranks. Requested {} ranks".format(
                    lib.max_concrete_rank, _ndim
                )
            )
        _dtype_str = "{}{}".format(_dtype, "*" * _ndim)

    _name = f"{_prefix}_{_dtype}_{_space}_{_layout}"
    _label = f"{_prefix}<{_dtype_str}, {_layout}, {_space}"

    # handle the trait argument
    if trait is not None:
        _trait = lib.get_memory_trait(trait)
        if _trait not in ("Managed", "Unmanaged"):
            _name = f"{_name}_{_trait}"
            _label = f"{_label}, {_trait}"

    # if fixed view
    if not dynamic:
        _name = f"{_name}_{_ndim}"

    # if a label was not provided
    if label is None:
        label = f"{_label}>"

    return getattr(lib, _name)(label if array is None else array, shape)


def unmanaged_array(*_args, **_kwargs):
    """This is kept for backwards compatibility"""
    return array(*_args, **_kwargs)


def create_mirror(src, copy=False):
    """Performs Kokkos::create_mirror"""
    return src.create_mirror(copy)


def create_mirror_view(src, copy=False):
    """Performs Kokkos::create_mirror_view"""
    return src.create_mirror_view(copy)


def deep_copy(dst, src):
    """Performs Kokkos::deep_copy"""
    return dst.deep_copy(src)



#!@PYTHON_EXECUTABLE@
# ************************************************************************
#
#                        Kokkos v. 3.0
#       Copyright (2020) National Technology & Engineering
#               Solutions of Sandia, LLC (NTESS).
#
# Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the Corporation nor the names of the
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY NTESS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NTESS OR THE
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Contact Christian R. Trott (crtrott@sandia.gov)
#
# ************************************************************************
#

from __future__ import absolute_import

__author__ = "Jonathan R. Madsen"
__copyright__ = (
    "Copyright 2020, National Technology & Engineering Solutions of Sandia, LLC (NTESS)"
)
__credits__ = ["Kokkos"]
__license__ = "BSD-3"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan R. Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"


import kokkos
import unittest

try:
    from . import _conftest as conf
except ImportError:
    import kokkos.test._conftest as conf


class PyKokkosBaseViewsTests(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        kokkos.initialize()

    @classmethod
    def tearDownClass(self):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def _print_info(self, _data):
        print(
            "[{}]> concrete type  : {}{}".format(
                self.shortDescription(), " " * 7, type(_data[0]).__name__
            )
        )
        print(
            "[{}]> dynamic type   : {}".format(
                self.shortDescription(), type(_data[1]).__name__
            )
        )

    def _print_data(self, _data, _zero, _idx):
        print(
            "[{}]> concrete zero  : {}".format(
                self.shortDescription(), _data[0].create_mirror_view()[_zero]
            )
        )
        print(
            "[{}]> concrete value : {}".format(
                self.shortDescription(), _data[0].create_mirror_view()[_idx]
            )
        )
        print(
            "[{}]> dynamic zero   : {}".format(
                self.shortDescription(), _data[1].create_mirror_view()[_zero]
            )
        )
        print(
            "[{}]> dynamic value  : {}".format(
                self.shortDescription(), _data[1].create_mirror_view()[_idx]
            )
        )

    def test_view_access(self):
        """view_access"""

        print("")
        for itr in conf.get_variants({"memory_traits": [kokkos.Atomic]}):
            _shape = itr[0]
            _idx = itr[1]
            _zeros = itr[2]
            _kwargs = itr[3]

            _data = conf.generate_variant(_shape, **_kwargs)

            self._print_info(_data)

            _host = [_data[0].create_mirror_view(), _data[1].create_mirror_view()]

            _host[0][_idx] = 1
            _host[1][_idx] = 2

            _data[0].deep_copy(_host[0])
            _data[1].deep_copy(_host[1])

            self._print_data(_data, _zeros, _idx)

            self.assertEqual(_data[0].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[1].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[0].create_mirror_view()[_idx], 1)
            self.assertEqual(_data[1].create_mirror_view()[_idx], 2)

    def test_view_iadd(self):
        """view_iadd"""

        print("")
        for itr in conf.get_variants({"memory_traits": [kokkos.Atomic]}):
            _shape = itr[0]
            _idx = itr[1]
            _zeros = itr[2]
            _kwargs = itr[3]

            _data = conf.generate_variant(_shape, **_kwargs)

            self._print_info(_data)

            _host = [_data[0].create_mirror_view(), _data[1].create_mirror_view()]

            _host[0][_idx] = 1
            _host[1][_idx] = 2

            _host[0][_idx] += 3
            _host[1][_idx] += 3

            _data[0].deep_copy(_host[0])
            _data[1].deep_copy(_host[1])

            self.assertEqual(_data[0].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[1].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[0].create_mirror_view()[_idx], 4)
            self.assertEqual(_data[1].create_mirror_view()[_idx], 5)

    def test_view_isub(self):
        """view_isub"""

        print("")
        for itr in conf.get_variants({"memory_traits": [kokkos.Atomic]}):
            _shape = itr[0]
            _idx = itr[1]
            _zeros = itr[2]
            _kwargs = itr[3]

            _data = conf.generate_variant(_shape, **_kwargs)

            self._print_info(_data)

            _host = [_data[0].create_mirror_view(), _data[1].create_mirror_view()]

            _host[0][_idx] = 10
            _host[1][_idx] = 20

            _host[0][_idx] -= 3
            _host[1][_idx] -= 3

            _data[0].deep_copy(_host[0])
            _data[1].deep_copy(_host[1])

            self.assertEqual(_data[0].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[1].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[0].create_mirror_view()[_idx], 7)
            self.assertEqual(_data[1].create_mirror_view()[_idx], 17)

    def test_view_imul(self):
        """view_imul"""

        print("")
        for itr in conf.get_variants({"memory_traits": [kokkos.Atomic]}):
            _shape = itr[0]
            _idx = itr[1]
            _zeros = itr[2]
            _kwargs = itr[3]

            _data = conf.generate_variant(_shape, **_kwargs)

            self._print_info(_data)

            _host = [_data[0].create_mirror_view(), _data[1].create_mirror_view()]

            _host[0][_idx] = 1
            _host[1][_idx] = 2

            _host[0][_idx] *= 3
            _host[1][_idx] *= 3

            _data[0].deep_copy(_host[0])
            _data[1].deep_copy(_host[1])

            self.assertEqual(_data[0].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[1].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[0].create_mirror_view()[_idx], 3)
            self.assertEqual(_data[1].create_mirror_view()[_idx], 6)

    #
    def test_view_create_mirror(self):
        """view_create_mirror"""
        print("")
        for itr in conf.get_variants(
            {"memory_traits": [kokkos.Atomic], "layouts": [kokkos.Unmanaged]}
        ):
            _shape = itr[0]
            _idx = itr[1]
            _zeros = itr[2]
            _kwargs = itr[3]

            _data = conf.generate_variant(_shape, **_kwargs)

            self._print_info(_data)

            _host = [_data[0].create_mirror_view(), _data[1].create_mirror_view()]

            _host[0][_idx] = 1
            _host[1][_idx] = 2

            _host[0][_idx] *= 3
            _host[1][_idx] *= 3

            _data[0].deep_copy(_host[0])
            _data[1].deep_copy(_host[1])

            self.assertEqual(_data[0].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[1].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[0].create_mirror_view()[_idx], 3)
            self.assertEqual(_data[1].create_mirror_view()[_idx], 6)

            _mirror_data = [
                kokkos.create_mirror(_data[0], copy=True),
                kokkos.create_mirror(_data[1], copy=True),
            ]

            self.assertEqual(_mirror_data[0][_zeros], 0)
            self.assertEqual(_mirror_data[1][_zeros], 0)
            self.assertEqual(_mirror_data[0][_idx], 3)
            self.assertEqual(_mirror_data[1][_idx], 6)

            _mirror_data = [
                kokkos.create_mirror(_data[0], copy=False),
                kokkos.create_mirror(_data[1], copy=False),
            ]

            self.assertEqual(_mirror_data[0][_zeros], 0)
            self.assertEqual(_mirror_data[1][_zeros], 0)
            self.assertNotEqual(_mirror_data[0][_idx], 3)
            self.assertNotEqual(_mirror_data[1][_idx], 6)

    #
    def test_view_create_mirror_view(self):
        """view_create_mirror_view"""

        print("")
        for itr in conf.get_variants({"memory_traits": [kokkos.Atomic]}):
            _shape = itr[0]
            _idx = itr[1]
            _zeros = itr[2]
            _kwargs = itr[3]

            _data = conf.generate_variant(_shape, **_kwargs)

            self._print_info(_data)

            _host = [_data[0].create_mirror_view(), _data[1].create_mirror_view()]

            _host[0][_idx] = 1
            _host[1][_idx] = 2

            _host[0][_idx] *= 3
            _host[1][_idx] *= 3

            _data[0].deep_copy(_host[0])
            _data[1].deep_copy(_host[1])

            self.assertEqual(_data[0].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[1].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[0].create_mirror_view()[_idx], 3)
            self.assertEqual(_data[1].create_mirror_view()[_idx], 6)

            _mirror_data = [
                kokkos.create_mirror_view(_data[0], copy=True),
                kokkos.create_mirror_view(_data[1], copy=True),
            ]

            self.assertEqual(_mirror_data[0][_zeros], 0)
            self.assertEqual(_mirror_data[1][_zeros], 0)
            self.assertEqual(_mirror_data[0][_idx], 3)
            self.assertEqual(_mirror_data[1][_idx], 6)

            _mirror_data = [
                kokkos.create_mirror_view(_data[0], copy=False),
                kokkos.create_mirror_view(_data[1], copy=False),
            ]

            self.assertEqual(_mirror_data[0][_zeros], 0)
            self.assertEqual(_mirror_data[1][_zeros], 0)

            if kokkos.get_host_accessible(_data[0].space):
                self.assertEqual(_mirror_data[0][_idx], 3)
            else:
                self.assertNotEqual(_mirror_data[0][_idx], 3)

            if kokkos.get_host_accessible(_data[1].space):
                self.assertEqual(_mirror_data[1][_idx], 6)
            else:
                self.assertNotEqual(_mirror_data[1][_idx], 6)

    #
    def test_view_deep_copy(self):
        """view_deep_copy"""
        print("")
        for itr in conf.get_variants(
            {"memory_traits": [kokkos.Atomic], "layouts": [kokkos.Unmanaged]}
        ):
            _shape = itr[0]
            _idx = itr[1]
            _zeros = itr[2]
            _kwargs = itr[3]
            if _kwargs["trait"] != kokkos.Managed:
                continue

            _data = conf.generate_variant(_shape, **_kwargs)

            self._print_info(_data)

            _host = [_data[0].create_mirror_view(), _data[1].create_mirror_view()]

            _host[0][_idx] = 1
            _host[1][_idx] = 2

            _host[0][_idx] *= 3
            _host[1][_idx] *= 3

            _data[0].deep_copy(_host[0])
            _data[1].deep_copy(_host[1])

            self.assertEqual(_data[0].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[1].create_mirror_view()[_zeros], 0)
            self.assertEqual(_data[0].create_mirror_view()[_idx], 3)
            self.assertEqual(_data[1].create_mirror_view()[_idx], 6)

            _copied_data = conf.generate_variant(_shape, **_kwargs)
            kokkos.deep_copy(_copied_data[0], _data[0])
            kokkos.deep_copy(_copied_data[1], _data[1])

            self.assertEqual(_copied_data[0].create_mirror_view()[_zeros], 0)
            self.assertEqual(_copied_data[1].create_mirror_view()[_zeros], 0)
            self.assertEqual(_copied_data[0].create_mirror_view()[_idx], 3)
            self.assertEqual(_copied_data[1].create_mirror_view()[_idx], 6)


# main runner
def run():
    # run all tests
    unittest.main()


if __name__ == "__main__":
    run()



#!/usr/bin/env python

import os
import sys
import argparse
import warnings
import platform
from skbuild import setup

# some Cray systems default to static libraries and the build
# will fail because BUILD_SHARED_LIBS will get set to off
if os.environ.get("CRAYPE_VERSION") is not None:
    os.environ["CRAYPE_LINK_TYPE"] = "dynamic"

cmake_args = [
    f"-DPYTHON_EXECUTABLE:FILEPATH={sys.executable}",
    f"-DPython3_EXECUTABLE:FILEPATH={sys.executable}",
    "-DCMAKE_INSTALL_RPATH_USE_LINK_PATH:BOOL=ON",
]

parser = argparse.ArgumentParser(add_help=False)
parser.add_argument("-h", "--help", help="Print help", action="store_true")


def set_cmake_bool_option(opt, enable_opt, disable_opt):
    global cmake_args
    try:
        if enable_opt:
            cmake_args.append(f"-D{opt}:BOOL=ON")
        if disable_opt:
            cmake_args.append(f"-D{opt}:BOOL=OFF")
    except Exception as e:
        print(f"Exception: {e}")


def add_arg_bool_option(lc_name, disp_name, default=None):
    global parser
    # enable option
    parser.add_argument(
        f"--enable-{lc_name}",
        action="store_true",
        default=default,
        help=f"Explicitly enable {disp_name} build",
    )
    # disable option
    parser.add_argument(
        f"--disable-{lc_name}",
        action="store_true",
        help=f"Explicitly disable {disp_name} build",
    )


# add options
add_arg_bool_option("experimental", "ENABLE_EXPERIMENTAL")
add_arg_bool_option("layouts", "ENABLE_LAYOUTS")
add_arg_bool_option("memory-traits", "ENABLE_MEMORY_TRAITS")
add_arg_bool_option("thin-lto", "ENABLE_THIN_LTO")
add_arg_bool_option("werror", "ENABLE_WERROR")
add_arg_bool_option("timing", "ENABLE_TIMING")
parser.add_argument(
    "--cxx-standard",
    default=14,
    type=int,
    choices=[14, 17, 20],
    help="Set C++ language standard",
)
parser.add_argument(
    "--enable-view-ranks",
    default=None,
    type=int,
    choices=[1, 2, 3, 4, 5, 6, 7],
    help="Maximum number of concrete view ranks",
)
parser.add_argument(
    "--kokkos-root",
    default=None,
    type=str,
    help="Path to kokkos install prefix",
)
parser.add_argument(
    "--cmake-args",
    default=[],
    type=str,
    nargs="*",
    help="{}{}{}".format(
        "Pass arguments to cmake. Use w/ pip installations",
        "and --install-option, e.g. --install-option=--cmake-args=",
        '"-DENABLE_LAYOUTS=ON -DKokkos_DIR=/usr/local/lib/cmake/Kokkos"',
    ),
)

args, left = parser.parse_known_args()
# if help was requested, print these options and then add '--help' back
# into arguments so that the skbuild/setuptools argparse catches it
if args.help:
    parser.print_help()
    left.append("--help")
sys.argv = sys.argv[:1] + left

set_cmake_bool_option(
    "ENABLE_EXPERIMENTAL", args.enable_experimental, args.disable_experimental
)
set_cmake_bool_option("ENABLE_LAYOUTS", args.enable_layouts, args.disable_layouts)
set_cmake_bool_option(
    "ENABLE_MEMORY_TRAITS",
    args.enable_memory_traits,
    args.disable_memory_traits,
)
set_cmake_bool_option("ENABLE_THIN_LTO", args.enable_thin_lto, args.disable_thin_lto)
set_cmake_bool_option("ENABLE_WERROR", args.enable_werror, args.disable_werror)
set_cmake_bool_option("ENABLE_TIMING", args.enable_timing, args.disable_timing)

cmake_args.append(f"-DCMAKE_CXX_STANDARD={args.cxx_standard}")

for itr in args.cmake_args:
    cmake_args += itr.split()

if args.enable_view_ranks is not None:
    cmake_args += [f"-DENABLE_VIEW_RANKS={args.enable_view_ranks}"]

if args.kokkos_root is not None:
    os.environ["CMAKE_PREFIX_PATH"] = ":".join(
        [args.kokkos_root, os.environ.get("CMAKE_PREFIX_PATH", "")]
    )

if platform.system() == "Darwin":
    # scikit-build will set this to 10.6 and C++ compiler check will fail
    darwin_version = platform.mac_ver()[0].split(".")
    darwin_version = ".".join([darwin_version[0], darwin_version[1]])
    cmake_args += [f"-DCMAKE_OSX_DEPLOYMENT_TARGET={darwin_version}"]

# DO THIS LAST!
# support PYKOKKOS_BASE_SETUP_ARGS environment variables because
#  --install-option for pip is a pain to use
# PYKOKKOS_BASE_SETUP_ARGS should be space-delimited set of cmake arguments, e.g.:
#   export PYKOKKOS_BASE_SETUP_ARGS="-DENABLE_LAYOUTS=OFF -DENABLE_MEMORY_TRAITS=ON"
env_cmake_args = os.environ.get("PYKOKKOS_BASE_SETUP_ARGS", None)
if env_cmake_args is not None:
    cmake_args += env_cmake_args.split(" ")


# --------------------------------------------------------------------------- #
#
def get_project_version():
    # open "VERSION"
    with open(os.path.join(os.getcwd(), "VERSION"), "r") as f:
        data = f.read().replace("\n", "")
    # make sure is string
    if isinstance(data, list) or isinstance(data, tuple):
        return data[0]
    else:
        return data


# --------------------------------------------------------------------------- #
#
def get_long_description():
    long_descript = ""
    try:
        long_descript = open("README.md").read()
    except Exception:
        long_descript = ""
    return long_descript


# --------------------------------------------------------------------------- #
#
def parse_requirements(fname="requirements.txt"):
    _req = []
    requirements = []
    # read in the initial set of requirements
    with open(fname, "r") as fp:
        _req = list(filter(bool, (line.strip() for line in fp)))
    # look for entries which read other files
    for itr in _req:
        if itr.startswith("-r "):
            # read another file
            for fitr in itr.split(" "):
                if os.path.exists(fitr):
                    requirements.extend(parse_requirements(fitr))
        else:
            # append package
            requirements.append(itr)
    # return the requirements
    return requirements


# suppress:
#  "setuptools_scm/git.py:68: UserWarning: "/.../<PACKAGE>"
#       is shallow and may cause errors"
# since 'error' in output causes CDash to interpret warning as error
with warnings.catch_warnings():
    print("CMake arguments: {}".format(" ".join(cmake_args)))
    setup(
        name="pykokkos-base",
        packages=["kokkos"],
        version=get_project_version(),
        cmake_args=cmake_args,
        cmake_languages=("C", "CXX"),
        long_description=get_long_description(),
        long_description_content_type="text/markdown",
        install_requires=parse_requirements("requirements.txt"),
        project_urls={"kokkos": "https://github.com/kokkos/kokkos"},
    )



#!/usr/bin/env python

import argparse
import numpy as np

#
# The python bindings for generate_view are in ex-generate.cpp
# The declaration and definition of generate_view are in user.hpp and user.cpp
# The generate_view function will return a Kokkos::View and will be converted
# to a numpy array
from ex_generate import generate_view, modify_view

#
# Importing this module is necessary to call kokkos init/finalize and
# import the python bindings to Kokkos::View which generate_view will
# return
#
import kokkos


def print_data(label, name, data):
    # write the type info
    print(
        "{:12} : {} (ndim={}, shape={})".format(
            label, type(data).__name__, data.ndim, data.shape
        )
    )

    # print the data
    if data.ndim == 1:
        for i in range(data.shape[0]):
            print("{:8}({}) = {}".format(name, i, data[i]))
    elif data.ndim == 2:
        for i in range(data.shape[0]):
            print(
                "{:8}({}) = [{}]".format(
                    name,
                    i,
                    " ".join("{}".format(data[i, j]) for j in range(data.shape[1])),
                )
            )
    else:
        raise ValueError("only 2 dimensions are supported")


def user_bindings(args):
    # get the kokkos view
    view = generate_view(args.ndim)
    print_data("Kokkos View", "view", view.create_mirror_view())

    # modify view (verify that casting works)
    modify_view(view)
    print_data("Modify View", "view", view.create_mirror_view())

    # wrap the buffer protocal as numpy array without copying the data
    arr = np.array(view.create_mirror_view(), copy=False)
    print_data("Numpy Array", "arr", arr)


def to_numpy(args):
    # get the kokkos view
    view = kokkos.array(
        "python_allocated_view",
        [args.ndim],
        dtype=kokkos.double,
        space=kokkos.DefaultHostMemorySpace,
    )

    for i in range(view.shape[0]):
        view[i] = i * (i % 2)
    print_data("Kokkos View", "view", view)

    # wrap the buffer protocal as numpy array without copying the data
    arr = np.array(view, copy=False)
    print_data("Numpy Array", "arr", arr)


def from_numpy(args):
    arr = np.ones([args.ndim, args.ndim], dtype=np.int32)
    for i in range(args.ndim):
        arr[i, i] = 0

    print_data("Numpy Array", "arr", arr)

    view = kokkos.array(arr, dtype=kokkos.int32, dynamic=True)
    print_data("Kokkos View", "view", view)


if __name__ == "__main__":
    try:
        kokkos.initialize()
        parser = argparse.ArgumentParser()
        parser.add_argument("-n", "--ndim", default=10, help="X dimension", type=int)
        args, argv = parser.parse_known_args()
        print("Executing to numpy...")
        to_numpy(args)
        print("Executing from numpy...")
        from_numpy(args)
        print("Executing user bindings...")
        user_bindings(args)
        kokkos.finalize()
    except Exception as e:
        import sys
        import traceback

        print(f"{e}")
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback)
        sys.exit(1)
